<?php

namespace Drupal\xray_audit\Plugin\xray_audit\tasks\QueryGroup;

use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Database\Query\Insert;
use Drupal\Core\Render\Markup;
use Drupal\Core\Url;
use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Plugin implementation of queries_data_node.
 *
 * @XrayAuditTaskPlugin (
 *   id = "queries_data_paragraphs",
 *   label = @Translation("Data about paragraphs"),
 *   description = @Translation("Reports on the Paragraphs of this site."),
 *   group = "queries_data",
 *   sort = 4,
 *   operations = {
 *      "paragraphs_count" = {
 *          "label" = "Paragraphs types and count",
 *          "description" = "Paragraphs types and count (default language)."
 *       },
 *      "paragraphs_count_by_hierarchy" = {
 *          "label" = "Paragraphs count grouped by type ordered hierarchically",
 *          "description" = "Paragraphs count grouped by type ordered
 *            hierarchically (default language)."
 *       },
 *      "paragraphs_count_by_language" = {
 *          "label" = "Paragraphs count grouped by language",
 *          "description" = "Paragraphs count grouped by language."
 *       },
 *    },
 *   dependencies = {"paragraphs"},
 *   batches = {
 *     "update-hierarchy-table" = "paragraphsHierarchyTableContentBatch"
 *   },
 *   install = "paragraphsInstallActions",
 *   uninstall = "paragraphsUninstallActions"
 * )
 */
class XrayAuditQueryTaskParagraphsPlugin extends XrayAuditQueryTaskPluginBase {

  const XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME = 'xra_paragraphs_hierarchy_tmp';

  /**
   * {@inheritdoc}
   */
  public function getDataOperationResult(string $operation = NULL) {
    switch ($operation) {
      case 'paragraphs_count':
        return $this->paragraphsTypes();

      case 'paragraphs_count_by_hierarchy':
        return $this->paragraphsTypesHierarchy();

      case 'paragraphs_count_by_language':
        return $this->paragraphsItemsCountByLanguage();
    }
    return [];
  }

  /**
   * Get data of Paragraph types.
   *
   * @return array
   *   Render array.
   */
  public function paragraphsTypes() {
    $headerTable = [
      $this->t('id'),
      $this->t('label'),
      $this->t('count'),
    ];
    $resultTable = [];

    $storage_type = $this->entityTypeManager->getStorage('paragraphs_type');
    $storage = $this->entityTypeManager->getStorage('paragraph');
    $language_by_default = $this->languageManager->getDefaultLanguage()->getId();
    $alias_count = 'count';

    $result = $storage
      ->getAggregateQuery()
      ->accessCheck(FALSE)
      ->aggregate('id', 'count', $language_by_default, $alias_count)
      ->condition('id', $this->getParagraphsUsedOnCurrentRevisions(), 'IN', $language_by_default)
      ->groupBy('type')
      ->execute();

    // Prepare results.
    $result_processed = [];
    foreach ($result as $row) {
      $result_processed[$row['type']] = $row['count'];
    }
    $total = 0;

    $paragraph_types = $storage_type->loadMultiple();
    foreach ($paragraph_types as $key => $paragraph_type) {
      $resultTable[] = [
        $key,
        $paragraph_type->label(),
        ($result_processed[$key] ?? 0),
      ];
      $total = $total + ($result_processed[$key] ?? 0);
    }

    $resultTable[] = [
      '',
      $this->t('Total'),
      $total,
    ];

    return [
      'header_table' => $headerTable,
      'results_table' => $resultTable,
    ];
  }

  /**
   * Get data of Paragraph types ordered hierarchically.
   *
   * @return array
   *   Render array.
   */
  public function paragraphsTypesHierarchy(): array {
    $this->paragraphsHierarchyCreateTemporaryTable();

    $headerTable = [
      $this->t('Level'),
      $this->t('Parent'),
      $this->t('Paragraph type'),
      $this->t('Count'),
    ];
    $language_by_default = $this->languageManager->getDefaultLanguage()->getId();

    $query = $this->database->select(self::XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME, 'tmp')
      ->fields('tmp', ['level', 'parent_bundle', 'type']);
    $query->addExpression('COUNT(tmp.id)', 'count');
    $results = $query->condition('tmp.langcode', $language_by_default)
      ->groupBy('tmp.hierarchy_id')
      ->groupBy('tmp.level')
      ->groupBy('tmp.type')
      ->groupBy('tmp.parent_bundle')
      ->orderBy('tmp.hierarchy_id')
      ->execute()
      ->fetchAll();

    // Prepare results.
    $resultTable = [];
    $paragraph_types = $this->entityTypeManager->getStorage('paragraphs_type')->loadMultiple();
    $total = 0;

    foreach ($results as $result) {
      $result = (array) $result;
      $indent_string = $result['level'] != '1' ? str_repeat(' ', $result['level'] - 1) . '- ' : NULL;
      $parent_label = $result['level'] == '1' ? $result['parent_bundle'] : NULL;
      $resultTable[] = [
        $result['level'],
        $parent_label,
        $indent_string . $paragraph_types[$result['type']]->label() . ' (' . $result['type'] . ')',
        $result['count'],
      ];
      $total = $total + $result['count'];
    }

    // Total row
    $resultTable[] = [
      '',
      '',
      Markup::create('<b>' . t('Total') . '</b>'),
      $total
    ];

    // Refresh row
    $url = Url::fromRoute('xray_audit.batch_process', ['task_plugin' => 'queries_data_paragraphs', 'batch_process' => "update-hierarchy-table"])->toString();
    $resultTable[] = [
      Markup::create('<a href="' . $url .'" class="button button--action button--primary">' . t('Refresh data') . '</a>'),
      '',
      ''
    ];

    return [
      'header_table' => $headerTable,
      'results_table' => $resultTable,
    ];
  }

  /**
   * Paragraphs items count by language.
   *
   * @return array
   *   Render array.
   */
  public function paragraphsItemsCountByLanguage() {

    $headerTable = [
      $this->t('label'),
      $this->t('type'),
      $this->t('langcode'),
      $this->t('total'),
    ];
    $resultTable = [];

    $storage_type = $this->entityTypeManager->getStorage('paragraphs_type');
    $storage = $this->entityTypeManager->getStorage('paragraph');
    $alias_count = 'count';

    // Get label of content types.
    $label = [];
    $types = $storage_type->loadMultiple();
    foreach ($types as $key => $type) {
      $label[$key] = $type->label();
    }

    $query = $storage->getAggregateQuery();
    $result = $query->accessCheck(FALSE)
      ->aggregate('id', 'COUNT', NULL, $alias_count)
      ->condition('id', $this->getParagraphsUsedOnCurrentRevisions(), 'IN')
      ->groupBy('langcode')
      ->groupBy('type')
      ->execute();

    foreach ($result as $row) {
      $resultTable[] = [
        'label' => $label[$row['type']],
        'id' => $row['type'],
        'langcode' => $row['langcode'],
        'total' => $row['count'] ?? 0,
      ];
    }

    return [
      'header_table' => $headerTable,
      'results_table' => $resultTable,
    ];

  }

  /**
   * Create/update the paragraphs Hierarchy temporally table.
   */
  protected function paragraphsHierarchyCreateTemporaryTable() {
    $state_service = \Drupal::service('state');
    $last_execution = $state_service->get('xray_audit.paragraps_hierarchy_temporary_table_creation', 0);

    if ($last_execution > strtotime("-7 day")) {
      return;
    }

    $url = Url::fromRoute('xray_audit.batch_process', ['task_plugin' => 'queries_data_paragraphs', 'batch_process' => "update-hierarchy-table"])->toString();
    $response = new RedirectResponse($url);
    $response->send();
  }

  /**
   * @return \Symfony\Component\HttpFoundation\RedirectResponse|null
   */
  public function paragraphsHierarchyTableContentBatch() {
    $this->database->truncate(self::XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME)->execute();
    $current_revision_paragraph_ids = self::getParagraphsUsedOnCurrentRevisions();

    $items = $this->database->select('paragraphs_item_field_data')
      ->fields('paragraphs_item_field_data', [
        'id',
        'parent_id',
        'type',
        'langcode',
        'parent_type',
      ])
      ->condition('id', $current_revision_paragraph_ids, 'IN')
      ->condition('parent_type', 'paragraph', '<>')
      ->execute()
      ->fetchAll();

    // Start a batch process.
    $operation_callback = [
      self::class,
      'paragraphsHierarchyOperationCallback',
    ];
    $finish_callback = [
      self::class,
      'paragraphsHierarchyFinishedCallback',
    ];

    // Define the messaging the user should see by default.
    $batch_builder = (new BatchBuilder())
      ->setTitle('Collecting information for the paragraph hierarchy report.')
      ->setFinishCallback($finish_callback)
      ->setInitMessage('The process in is initiating.')
      ->setProgressMessage('The process is executing.')
      ->setErrorMessage('The batch process failed.');
    $batch_builder->addOperation(
      $operation_callback,
      [$items, $current_revision_paragraph_ids]
    );
    batch_set($batch_builder->toArray());

    $url = Url::fromRoute('xray_audit.task_page', [
      'group_plugin' => 'queries_data',
      'task_plugin_operation' => 'queries_data_paragraphs-paragraphs_count_by_hierarchy'
    ])->toString();

    return batch_process($url);
  }

  /**
   * @param array $items
   * @param $context
   *
   * @return void
   */
  public static function paragraphsHierarchyOperationCallback(array $items, array $current_revision_paragraph_ids, &$context) {
    $length = 100;

    if (!isset($context['sandbox']['progress'])) {
      $context['sandbox']['progress'] = 0;
      $context['sandbox']['errors'] = [];
      $context['sandbox']['max'] = count($items);
    }

    $query = \Drupal::service('database')->insert(self::XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME)
      ->fields([
        'hierarchy_id',
        'id',
        'parent_id',
        'type',
        'parent_type',
        'parent_bundle',
        'langcode',
        'level',
      ]);

    foreach (array_slice($items, $context['sandbox']['progress'], $length, TRUE) as $result) {
      $result = (array) $result;
      self::paragraphsHierarchyProcessLevel($result, $query, $current_revision_paragraph_ids);
      $context['sandbox']['progress']++;
    }
    $query->execute();

    $context['finished'] = $context['sandbox']['progress'] / $context['sandbox']['max'];
    $context['message'] = t('Processed @current paragraphs of @total', [
      '@current' => $context['sandbox']['progress'],
      '@total' => $context['sandbox']['max'],
    ]);

  }

  /**
   * @param $success
   * @param array $results
   * @param array $operations
   *
   * @return void
   */
  public static function paragraphsHierarchyFinishedCallback($success, array $results, array $operations) {
    $state_service = \Drupal::service('state');
    $state_service->set('xray_audit.paragraps_hierarchy_temporary_table_creation', time());
  }

  /**
   * @param array $result
   * @param \Drupal\Core\Database\Query\Insert $query
   *
   * @return void
   */
  public static function paragraphsHierarchyProcessLevel(array $result, Insert $query, array $current_revision_paragraph_ids): void {
    $result = (array) $result;
    $entity = NUll;
    $result['status'] = 1;
    $entity_type_manager = \Drupal::service('entity_type.manager');

    if ($result['parent_type'] == 'node') {
      $entity = $entity_type_manager->getStorage('node')->load($result['parent_id']);
    }
    if ($result['parent_type'] == 'block_content') {
      $entity = $entity_type_manager->getStorage('block_content')->load($result['parent_id']);
    }
    if (!$entity) {
      return;
    }
    $result['status'] = $entity->isPublished();

    // Check if the paragraph is referred in the node.
    foreach ($entity->getFields() as $field) {
      $definition = $field->getFieldDefinition();
      if ($definition->getType() === 'entity_reference_revisions'
        && $definition->getSetting('target_type') === 'paragraph'
        && array_search($result['id'], array_column($field->getValue(), 'target_id')) == FALSE) {
        return;
      }
    }

    $result['hierarchy_id'] = $result['parent_type'] . "-" . $result['type'];
    $result['parent_bundle'] = $result['parent_type'];
    $result['level'] = 1;
    $query->values($result);
    self::paragraphsHierarchyProcessChildLevel($result, $query, $current_revision_paragraph_ids);
  }

  /**
   * Get the list of paragraphs that are being used on current revisions.
   *
   * This is used to not show unused paragraphs in paragraphs hierarchy.
   *
   * @return array
   *   IDs of unused paragraphs.
   */
  protected function getParagraphsUsedOnCurrentRevisions() {
    $paragraph_field_tables_query = $this->database->select('paragraphs_item_field_data', 'pd');
    $paragraph_field_tables_query->addExpression('CONCAT(pd.parent_type, :underscores , pd.parent_field_name)', 'table_name', [':underscores' => '__']);
    $paragraph_field_tables_query->groupBy('pd.parent_type');
    $paragraph_field_tables_query->groupBy('pd.parent_field_name');
    $paragraph_field_tables = $paragraph_field_tables_query->execute()->fetchAllKeyed(0, 0);
    $paragraphs_ids = [];
    foreach ($paragraph_field_tables as $paragraph_field_table) {
      [, $paragraph_field_table_field_name] = explode('__', $paragraph_field_table);
      if ($this->database->schema()->tableExists($paragraph_field_table)) {
        $paragraphs_ids = array_merge($paragraphs_ids, $this->database->select($paragraph_field_table)
          ->fields($paragraph_field_table, [$paragraph_field_table_field_name . '_target_id'])
          ->execute()
          ->fetchAllKeyed(0, 0));
      }
    }
    return $paragraphs_ids;
  }

  /**
   * Process a child level of the paragraphs Hierarchy temporally table.
   *
   * @param array $parent
   *   Parent result.
   * @param \Drupal\Core\Database\Query\Insert $query
   *   Drupal Query object.
   */
  public static function paragraphsHierarchyProcessChildLevel(array $parent, Insert $query, array $current_revision_paragraph_ids): void {
    $results = \Drupal::service('database')->select('paragraphs_item_field_data')
      ->fields('paragraphs_item_field_data', [
        'id',
        'parent_id',
        'type',
        'langcode',
        'parent_type',
      ])
      ->condition('id', $current_revision_paragraph_ids, 'IN')
      ->condition('parent_id', $parent['id'])
      ->condition('parent_type', 'paragraph')
      ->execute()
      ->fetchAll();

    foreach ($results as $result) {
      $result = (array) $result;
      $result['hierarchy_id'] = $parent['hierarchy_id'] . "-" . $result['type'];
      $result['parent_bundle'] = $parent['type'];
      $result['level'] = $parent['level'] + 1;
      $query->values($result);

      if ($result['level'] <= 10) {
        self::paragraphsHierarchyProcessChildLevel($result, $query, $current_revision_paragraph_ids);
      }
    }
  }

  /**
   * Get the table configuration for the paragraphs Hierarchy temporally table.
   *
   * @return array
   *   Table configuration array.
   */
  protected function paragraphsHierarchyGetTableConfiguration(): array {
    return [
      'description' => 'Temporary table to get paragraph usage with hierarchy',
      'fields' => [
        'hierarchy_id' => [
          'type' => 'varchar',
          'length' => 255,
        ],
        'id' => [
          'type' => 'int',
          'not null' => TRUE,
          'default' => '0',
          'unsigned' => TRUE,
        ],
        'parent_id' => [
          'type' => 'int',
          'not null' => TRUE,
          'default' => '0',
          'unsigned' => TRUE,
        ],
        'type' => [
          'type' => 'varchar',
          'length' => 32,
        ],
        'parent_type' => [
          'type' => 'varchar',
          'length' => 32,
        ],
        'parent_bundle' => [
          'type' => 'varchar',
          'length' => 32,
        ],
        'langcode' => [
          'type' => 'varchar',
          'length' => 10,
        ],
        'level' => [
          'type' => 'int',
          'not null' => TRUE,
          'default' => '1',
        ],
      ],
    ];
  }

  /**
   * Create the requirements for the task.
   */
  public function paragraphsInstallActions(): void {
    if (!$this->database->schema()->tableExists(self::XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME)) {
      $this->database->schema()->createTable(self::XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME, $this->paragraphsHierarchyGetTableConfiguration());
    }
  }

  /**
   * Uninstall the requirements for the task.
   */
  public function paragraphsUninstallActions(): void {
    $this->database->schema()->dropTable(self::XRA_PARAGRAPH_TEMPORARY_TAMBLE_NAME);
  }

}
